#!/usr/bin/perl

# RoadHog Search CGI version 0.9
# (c) 2001 Frederik Ramm
# http://www.remote.org/frederik/roadhog/software/
# released under GPL

use String::Similarity;
use CGI;

$maxdist = 10;
$query = CGI::new ();

if (!(defined ($query->param ("place")) || defined ($query->param ("coord"))))
{
	header ("Search");
	print "You can search either for a place name or for co-ordinates. The\n";
	print "place names come from a database I've created manually, so not\n";
	print "<em>every</em> name will be there, but I have entered the names\n";
	print "of any villages, many farms and crofts, forstests, rivers, glens,\n";
	print "lochs, and mountains I passed. <p>\n";
	print "The script will also find similar-sounding names (which is essential\n";
	print "around here since every map spells those ex-Gaelic names differently!).";
	print "<p>";
	searchform ();
	footer ("");
}
else
{
	# form filled out
	$size = $query->param ("size");
	if ($size eq "")
	{
		$size = "small";
	}
	
	if ($query->param ("place") ne "")
	{
		$orig_place = lc ($query->param ("place"));
		$place = lc ($query->param ("place"));
		open (PLACES, "control/places.txt") || trouble ("places.txt");
		while (<PLACES>)
		{
			chomp;
			last if (/^\s*$/);
			$route[$count++] = $_;
		}
		$minsim = 0.6;
		while (<PLACES>)
		{
			($word, $qualifier, $routes) = split(/:/);
			if ($word =~ /(.*)\*(.*)/)
			{
				$sim1 = similarity $place, lc ($1), $minsim;
				$sim2 = similarity $place, lc ("$1 $2"), $minsim;
				$sim = ($sim1 > $sim2) ? $sim1 : $sim2;
			}
			else
			{
				$sim = similarity $place, lc ($word), $minsim;
			}
			if ($sim >= $minsim)
			{
				$key = $word . "_" . $qualifier;
				$match{$key}{"sim"} = $sim;
				$match{$key}{"ptr"} = $routes;
				$found++;
				$minsim = 1 if ($sim == 1)  
			}
		}
		close (PLACES);

		if ($found == 0)
		{
			header ("Nothing found");
			print "No places matching your input were found.<hr><h3>Try again:</h3>";
			searchform ();
		}
		else
		{ 
			header ("Search Results");
			if ($minsim == 1)
			{
				print "Here's a list of routes that run through, past, along, or near the ";
				print "place you entered.\n";
			}
			else
			{
				print "Here's a list of routes that run through, past, along, or near some ";
				print "places that sound similar to what you entered.\n";
			}
			print "Note that this doesn't necessarily mean that the ";
			print "place/feature you are looking for is actually visible on one of the photos!\n";
			
			print "<dl>\n";
			@sorted = sort ( { $match{$b}{"sim"} <=> $match{$a}{"sim"} } keys (%match));
			$lastsim = 0;
			$outcount = 0;
			foreach $key (@sorted)
			{
				if ($match{$key}{"sim"} < $minsim)
				{
					last;
				}
				if (($match{$key}{"sim"} < $lastsim) && ($outcount > 5))
				{
					last;
				}
				$lastsim = $match{$key}{"sim"};
				$outcount++;
				($word, $qualifier) = split (/_/, $key);
				print "<dt>Routes at/in/through/past <strong>";
				$word =~ s/\*/ /g;
				print $word;
				print " ($qualifier)" if ($qualifier ne "");
				print ":</strong></dt>\n<dd>";
				
				foreach $route (split /,/, $match{$key}{"ptr"})
				{
					($url, $text) = split (/:/, $route[$route]);
					print "<a href =\"roadhog.cgi?show=$url/$size/1\">$text</a>";
					print "\n<br>";
				}
			}
			print ("</dl>\n");
		}
		print ("<p>End of search results.");
		footer ("");	
	}
	elsif ($query->param ("coord") ne "")
	{
		$entered = $query->param ("coord");
		$use_osgrid = 0;
		if ($entered =~ /\s*N([BCDGHIMNO])\D*(\d*)\D+(\d*)/i)
		{
			$gridy = 500000;
			($square, $east, $north) = ($1, $2, $3);
			$pos = index ("VWXYZQRSTULMNOPFGHJKABCDE", $square);
			$gridx += ($pos % 5) * 100000;
			$gridy += int ($pos / 5) * 100000;
			$easting = substr ($east."00000", 0, 5);
			$northing = substr ($north."00000", 0, 5);
			$gridx += $easting;
			$gridy += $northing;
			$use_osgrid = 1;
		}
		if ($entered =~ /(\S{2,})\s+(\S{2,})/)
		{
			($lat, $lon) = ($1, $2);
		}
		elsif ($entered =~ /(\S{2,}?)\s*[,;:\/\-]\s*(\S{2,})/)
		{
			($lat, $lon) = ($1, $2);
		}
		elsif ($entered =~ /(\S{2,})([EW]\S+)/i)
		{
			($lat, $lon) = ($1, $2);
		}
		else
		{
			$error = 1;
		}
	
		if (($error != 1) && ($use_osgrid != 1))
		{
			$latns = 1; $lonew = 1;
			if ($lat =~ /(.*)([NS])(.*)/i)
			{
				$lat = "$1$3";
				$latns = (lc($2) eq "n") ? 1 : -1;
			}
			elsif ($lat =~ /(.*)-(.*)/)
			{
				$lat = "$1$2";
				$latns = -1;
			}
			if ($lon =~ /(.*)([EW])(.*)/i)
			{
				$lon = "$1$3";
				$lonew = (lc($2) eq "e") ? 1 : -1;
			}
			elsif ($lon =~ /(.*)-(.*)/)
			{
				$lon = "$1$2";
				$lonew = -1;
			}

			$lat = parsecoord ($lat, $latns);
			$lon = parsecoord ($lon, $lonew);

			if (($lat < -90) || ($lat > 90) || ($lon < -180) || ($lon > 180))
			{
				$error = 1;
			}
		}
	
		if ($error == 1)
		{
			header ("Bad Location");
			print <<EOF;
			The co-ordinates you entered could not be understood. Your input will
			be understood if you follow these hints:

			<h3>For Latitude/Longitude pairs (WGS-84 map datum)</h3>
			<ul>
			<li>enter latitude first, then longitude;
			<li>separate latitude and longitude by a space or comma
			<li>do not put a space or comma in your latitude or longitude
			<li>Enter latitude and longitude...<ul>
			<li>... either as degrees, minutes, and seconds,
				using any character but a decimal point as a separator: 5723'12"
			<li>... or as degrees and minutes, where minutes may have a decimal 
				point and a fraction: 5723.123'
			<li>... or as degrees with a fraction: 57.9876</ul>
			<li>Use a letter (N,S with latitude; E,W with longitude) to denote the 
				hemisphere you're talking about. You can also use signs with the
				degree specification - a negative latitude is south, and a negative
				longitude is west.
			</ul>
			A valid example for a full co-ordinate specification is: 
			<em>N5748.10' W00540.32'</em>. The specification 
			<em>57.80166,-5.672</em> is equally valid and describes (approximately)
			the same location.

			<h3>For British Grid (a.k.a. "Landranger Grid") co-ordinates</h3>
			<ul>
			<li>Enter the two-letter square designator first, followed by 
			    the "easting", a separator, and the "northing". 
		    <li>Any non-digit is permitted as a separator.
			<li>If you give easting and northing with less than five
			    digit precision (usual Landranger co-ordinates will be
				given with three digits), zeroes will be added to
				make five-digit numbers. 
		    <li>Co-ordinates outside the square designators
			    NB, NC, ND, NG, NH, NI, NM, NN, NO are not recognized
				(and not covered either).
			</ul>
			A valid grid specification is "NG 81837 85005". You could
			also write it, a bit less precisely, as "NG818,850". Both
			describe roughly the same point as the latitude/longitude
			pair given further up.
			<hr>
			<h3>Try again:</h3>
EOF
			searchform ();
			footer ("");
			exit 1;
		}

		header ("Location search result");

		if ($use_osgrid)
		{
			print "<table><tr><td>OSGS grid:<td>";
			print "N$square";
			printf ("%03d %03d", $easting/100, $northing/100);
			print "<tr><td><td>($gridx East, $gridy North)";
			print "</table>\n";
		}
		else
		{
			print "<table cellpadding=0 cellspacing=0><tr><td>Latitude:</td><td align=right>";
			$ll = $lat;
			if ($ll < 0)
			{
				print "S";
				$ll = -$ll;
			}
			else
			{
				print "N";
			}
			printf ("%02d", int ($ll));	
			$ll -= int ($ll);
			$ll *= 60;
			printf ("%02d'", int ($ll));	
			$ll -= int ($ll);
			$ll *= 60;
			printf ("%02d\"", int ($ll));	
			printf "</tr><td align=right>(%9.6f)</td>", $lat;

			print "<tr><td>Longitude:</td><td align=right>";
			$ll = $lon;
			if ($ll < 0)
			{
				print "W";
				$ll = -$ll;
			}
			else
			{
				print "E";
			}
			printf ("%03d", int ($ll));	
			$ll -= int ($ll);
			$ll *= 60;
			printf ("%02d'", int ($ll));	
			$ll -= int ($ll);
			$ll *= 60;
			printf ("%02d\"", int ($ll));	
			printf "</td><td align=right>(%9.6f)</td></table>", $lon;
		}

		print ("<p>");
		
		open (LOC, "control/location.txt") || trouble ("location.txt");
		while (<LOC>)
		{
			chomp;
			last if (/^\s*$/);
			$dist{$count} = 2000 * $maxdist;
			$route[$count++] = $_;
		}
		$_= <LOC>;
		($minlong, $maxlat, $longfactor, $latfactor) = split (/:/);
		$x = ($lon - $minlong) * $longfactor;
		$y = ($maxlat - $lat) * $latfactor;

		if ($use_osgrid)
		{
			$x = $gridx;
			$y = $gridy;
		}

		$miny = $y - $maxdist * 1000;
		$maxy = $y + $maxdist * 1000;
		$minx = $x - $maxdist * 1000;
		$maxx = $x + $maxdist * 1000;

		while (<LOC>)
		{
			if ($use_osgrid)
			{
				($dummy, $dummy, $px, $py, $route) = split(/:/);
			} 
			else
			{
				($px, $py, $dummy, $dummy, $route) = split(/:/);
			}
			next if ($px < $minx);
			next if ($py < $miny);
			next if ($px > $maxx);
			next if ($py > $maxy);

			$dist = sqrt ( ($px-$x)**2 + ($py-$y)**2 );
			if (!defined ($dist{$route}) || $dist{$route} > $dist)
			{
				$dist{$route} = $dist;
			}
			$found++;
		}
		close (LOC);

		if ($found == 0)
		{
			print "There are no routes that lie within $maxdist kilometres of ";
			print "the location you entered.\n";
			print "<hr><h3>Try again:</h3>\n";
			searchform ();
			footer ("");
			exit 1;
		}

		print "<dl>";

		foreach $index (sort { $dist{$a} <=> $dist{$b} } keys (%dist))
		{
			if ($dist{$index} < 500)
			{
				$group = "500 metres";
			}
			elsif ($dist{$index} < 2000)
			{
				$group = "2 kilometres";
			}
			elsif ($dist{$index} < 5000)
			{
				$group = "5 kilometres";
			}
			else
			{
				$group = sprintf ("%d kilometres", $maxdist);
				last if ($dist{$index} > $maxdist * 1000);
			}	
			if ($lastgroup ne $group)
			{
				if ($lastgroup ne "")
				{
					print "<br>\n<dt>Additionally, t";
					$add = "";
				}
				else
				{
					print "<dt>T";
					$add = "of the co-ordinates you entered";
				}
				print "he following routes contain at least one point within $group";
				print "$add:\n<dd>\n";
				$lastgroup = $group;
			}

			($url, $text) = split (/:/, $route[$index]);
			print "<a href =\"roadhog.cgi?show=$url/$size/1\">$text</a>";
			if ($dist{$index} < 1000)
			{
				printf " <em>(%3d metres)</em>", int ($dist{$index} / 50 + 0.5) * 50;
			}
			else
			{
				printf " <em>(%3.1fkm)</em>", $dist{$index} / 1000;
			}
			print "\n<br>";
		}
		
		print ("</dl>");
		print ("<p>End of search results.");
		footer ("");

	}
}
print "\n";
			
sub parsecoord 
{
	my ($coor, $fac) = @_;
	if ($coor =~ /(\d+)[^0-9.]+(\d+(\.\d+)?)/)
	{
		$val = $1 + ($2 / 60);
	}
	elsif ($coor =~ /(\d+)\D+(\d+)(\D+(\d+))/)
	{
		$val = $1 + ($2 / 60) + ($4 / 3600);
	}
	elsif ($coor =~ /([0-9.]+)/)
	{
		$val = $1;
	}
	else
	{	
		$val = -1000;
	}
	$val  *= $fac;
	return ($val);
}
	
sub header {
	my ($title) = @_;
	print <<EOF;
Content-type: text/html

<html>
<head><title>On the Road in North-West Scotland - $title</title>
<style type="text/css">
<!--
body { margin: 2%; background: #ffffee; color: black; }
p,td,th,h1,h2,h3,ul,ol,li,b,em,strong.dt,dd,dl { font-family: "New Century Schoolbook", Times, serif; color: black; }
a:visited { color: #551A8B; }
a:link { color: #0000EE; }
//-->
</style>

</head>
<body bgcolor=white text=black link=blue>
<table width=100% cellpadding=5><tr bgcolor="#bbbbff"><td><font size=+2><b>$title</b></font></td></tr></table>

<p>
<table width=100%><tr><td>
EOF
}

sub footer {
	print "</table>";
	print "<table width=100%><tr bgcolor=\"#ccccaa\"><td><table width=100% cellpadding=0 cellpadding=0><tr><td><a href=\"index.html\">Place List</a> &nbsp; <a href=\"search.cgi\">Search</a> &nbsp; <a href=\"faq.html\">Help &amp; FAQ</a></td><td align=right>v0.9</td></tr></table></td></tr></table>";
	print "</body></html>\n";
}

sub searchform {

	print <<EOF;
	<table>
	<tr>
	<td>
	<form method=post action="search.cgi">
	Enter a place name:
	<td>
	<input type=text name=place>
	<td>
	<input type=submit>
	<tr>
	<td>
	<td colspan=2>
	<font size=-1><em>(Examples: "Dundonnell", "Loch Maree", "Glen Docherty")
		</em></font>
	<tr>
	<td colspan=3 align=center>
	</form><br>
	--- OR ---
	<form method=post action="search.cgi">
	<tr>
	<td>
	Enter a location:
	<td>
	<input type=text name=coord>
	<td>
	<input type=submit>
	<tr>
	<td>
	<td colspan=2>
	<font size=-1><em>(Examples: "N55.1234,W5.9876", "N5748.10' W00540.32'",
	    "NG 739,711" -
		most formats magically recognized, either lat/long according to WGS-84
		or grid values according to British National Grid)</em></font>
	</table>
EOF

}

sub trouble ()
{
	my ($msg) = @_;
	header ("Problem");
	print "<p>The search script has encountered a technical problem. If this ";
	print "condition persists, please contact the author at frederik\@remote.org.<p>";
	print "Additional fault information: <tt>$msg</tt><p>" if ($msg ne "");
	footer ("");
	exit (1);
}
	
